#include "config.h"
#include <dirent.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "configio.h"
#include "defines.h"
#include "errcodes.h"
#include "fdbuf.h"
#include "itoa.h"
#include "list.h"

typedef list<mystring> slist;

#define fail(MSG) do { ferr << MSG << endl; return false; } while(0)
#define fail_sys(MSG) do{ ferr << MSG << strerror(errno) << endl; return false; }while(0)

struct remote
{
  mystring host;
  mystring proto;
  slist options;
  remote(const mystring& h, const mystring& p, const mystring& o);
  ~remote();
};

remote::remote(const mystring& h, const mystring& p, const mystring& o)
  : host(h), proto(p)
{
  if(!!o) {
    int start = 0;
    int end = o.find(' ');
    while(end > start) {
      options.append(o.sub(start, end-start));
      start = end+1;
      end = o.find(' ', start);
    }
    options.append(o.right(start));
  }
}

remote::~remote() { }

typedef list<remote> rlist;

static rlist remotes;
static int pausetime = 60;
static bool reload_config = false;

void catch_hup(int)
{
  signal(SIGHUP, catch_hup);
  reload_config = true;
}

bool load_remotes()
{
  slist rtmp;
  if(!config_readlist("remotes", rtmp) ||
     rtmp.count() == 0)
    return false;
  remotes.empty();
  for(slist::iter r(rtmp); r; r++) {
    int i = (*r).find(' ');
    if(i < 0)
      remotes.append(remote((*r), "smtp", 0));
    else {
      mystring host = (*r).left(i);
      mystring rem = (*r).right(i+1);
      i = rem.find(' ');
      if(i < 0)
	remotes.append(remote(host, rem, 0));
      else
	remotes.append(remote(host, rem.left(i), rem.right(i+1)));
    }
  }
  return remotes.count() > 0;
}

bool load_config()
{
  bool result = true;
  if(!load_remotes()) result = false;
  
  if(!config_readint("pausetime", pausetime)) {
    pausetime = 60;
    result = false;
  }
  return result;
}

static slist files;
static bool reload_files = false;

void catch_alrm(int)
{
  signal(SIGALRM, catch_alrm);
  reload_files = true;
}

bool load_files()
{
  static time_t mtime = 0;
  struct stat sb;
  if(stat(".", &sb) == -1)
    fail_sys("Cannot stat queue directory");
  if(mtime == sb.st_mtime)
    return true;
  fout << "Rescanning queue." << endl;
  DIR* dir = opendir(".");
  if(!dir)
    fail_sys("Cannot open queue directory: ");
  files.empty();
  struct dirent* entry;
  while((entry = readdir(dir)) != 0) {
    const char* name = entry->d_name;
    if(name[0] == '.')
      continue;
    files.append(name);
  }
  closedir(dir);
  mtime = sb.st_mtime;
  return true;
}

void exec_protocol(mystring filename, remote& remote)
{
  if(close(0) == -1 || close(1) == -1 || close(2) == -1)
    return;
  mystring program = PROTOCOL_DIR + remote.proto;
  const char* args[4+remote.options.count()];
  unsigned i = 0;
  args[i++] = program.c_str();
  for(slist::iter opt(remote.options); opt; opt++)
    args[i++] = (*opt).c_str();
  args[i++] = remote.host.c_str();
  args[i++] = filename.c_str();
  args[i++] = 0;
  execv(args[0], (char**)args);
}

#undef fail
#define fail(MSG) do { fout << MSG << endl; return false; } while(0)
#define fail2(MSG1,MSG2) do{ fout << MSG1 << MSG2 << endl; return false; }while(0)

bool catchsender(pid_t pid)
{
  int status;
  if(waitpid(pid, &status, 0) == -1)
    fail("Error catching the child process return value.");
  else {
    if(WIFEXITED(status)) {
      status = WEXITSTATUS(status);
      if(status)
	fail2("Sending failed: ", errorstr[status]);
      else {
	fout << "Sent file." << endl;
	return true;
      }
    }
    else
      fail("Sending process crashed or was killed.");
  }
}

bool send_one(mystring filename, remote& remote)
{
  fout << "Starting delivery: protocol: " << remote.proto
       << " host: " << remote.host
       << " file: " << filename << endl;
  pid_t pid = fork();
  switch(pid) {
  case -1:
    fail("Fork failed.");
  case 0:
    exec_protocol(filename, remote);
    exit(ERR_EXEC_FAILED);
  default:
    if(!catchsender(pid))
      return false;
    if(unlink(filename.c_str()) == -1)
      fail("Can't unlink file.");
  }
  return true;
}

bool send_all()
{
  if(remotes.count() <= 0)
    fail("No remote hosts listed for delivery");
  if(files.count() == 0)
    return true;
  fout << "Starting delivery, "
       << itoa(files.count()) << " message(s) in queue." << endl;
  for(rlist::iter remote(remotes); remote; remote++) {
    for(slist::iter file(files); file; files.remove(file)) {
      if(!send_one(*file, *remote))
	break;
    }
  }
  fout << "Delivery complete, "
       << itoa(files.count()) << " message(s) remain." << endl;
  return true;
}

static int trigger;

bool open_trigger()
{
  trigger = open(QUEUE_TRIGGER, O_RDONLY|O_NONBLOCK);
  if(trigger == -1)
    fail("Could not open trigger file.");
  return true;
}

bool read_trigger()
{
  if(trigger != -1) {
    char buf[1024];
    read(trigger, buf, sizeof buf);
    close(trigger);
  }
  return open_trigger();
}

bool do_select()
{
  fd_set readfds;
  FD_ZERO(&readfds);
  FD_SET(trigger, &readfds);
  struct timeval timeout;
  timeout.tv_sec = pausetime;
  timeout.tv_usec = 0;
  int s = select(trigger+1, &readfds, 0, 0,
		 (files.count() == 0) ? 0 : &timeout);
  if(s == 1) {
    fout << "Trigger pulled." << endl;
    read_trigger();
    reload_files = 1;
  }
  else if(s == -1 && errno != EINTR)
    fail("Internal error in select.");
  if(reload_config) {
    fout << "Signal received, rescanning control files." << endl;
    load_config();
  }
  if(reload_files)
    load_files();
  return true;
}

int main(int, char*[])
{
  if(!open_trigger())
    return 1;
  if(chdir(QUEUE_MSG_DIR) == -1) {
    fout << "Could not chdir to queue message directory." << endl;
    return 1;
  }
  
  signal(SIGALRM, catch_alrm);
  signal(SIGHUP, catch_hup);
  load_config();
  load_files();
  for(;;) {
    send_all();
    do_select();
  }
  return 0;
}
